use crate::table::raw::RawTable;

use crate::environment::Txn;
use anyhow::anyhow;
use byte_lamination::ByteLamination;
use libmdbx::{TransactionKind, RW};
use std::borrow::{Borrow, Cow};
use std::marker::PhantomData;

pub struct Table<K, V> {
    pub raw: RawTable,
    pub _marker: PhantomData<(K, V)>,
}

impl<K: for<'a> ByteLamination<'a>, V: for<'a> ByteLamination<'a>> Table<K, V> {
    pub fn get<'txn, 'qk, QK: ByteLaminationHelper<'qk, K>, TK: TransactionKind>(
        &self,
        txn: &Txn<'txn, TK>,
        k: QK,
    ) -> anyhow::Result<Option<V>> {
        let k_bytes = k.as_cow_bytes_helper();

        let val_opt = self.raw.get(txn, &k_bytes)?;

        match val_opt {
            Some(cow_bytes) => {
                let value = V::try_from_bytes(cow_bytes)
                    .map_err(|e| anyhow!("failed to convert value bytes to value: {e:?}"))?;
                Ok(Some(value))
            }
            None => Ok(None),
        }
    }

    pub fn put<
        'txn,
        'qk,
        'qv,
        QK: ByteLaminationHelper<'qk, K>,
        QV: ByteLaminationHelper<'qv, V>,
    >(
        &self,
        txn: &Txn<'txn, RW>,
        k: QK,
        v: QV,
    ) -> anyhow::Result<()> {
        let k_bytes = k.as_cow_bytes_helper();
        let v_bytes = v.as_cow_bytes_helper();
        self.raw.put(txn, &k_bytes, &v_bytes)?;
        Ok(())
    }

    /// Returns true if the key was present.
    pub fn delete<'txn>(
        &self,
        txn: &Txn<'txn, RW>,
        k: &(impl Borrow<K> + ?Sized),
    ) -> anyhow::Result<bool> {
        let k_bytes = k.borrow().as_cow_bytes();

        self.raw.delete(txn, &k_bytes)
    }
}

/// `T`: a ByteLamination that this type is equivalent to.
pub trait ByteLaminationHelper<'a, T: ByteLamination<'a>> {
    fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]>;
}

impl<'a, T: ByteLamination<'a>> ByteLaminationHelper<'a, T> for T {
    fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]> {
        self.as_cow_bytes()
    }
}

impl<'a, 'b, T: ByteLamination<'a>> ByteLaminationHelper<'a, T> for &'b T {
    fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]> {
        self.as_cow_bytes()
    }
}

// can't get rust happy with this...
// impl<'a, T: ?Sized + ToOwned> ByteLaminationHelper<'a, Cow<'a, T>> for &'a T
//     where for<'b> Cow<'b, T>: ByteLamination<'b> {
//     fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]> {
//         let x: Cow<'a, T> = Cow::Borrowed(self);
//         let y: &'a Cow<'a, T> = &x;
//         Cow::<'_, T>::as_cow_bytes(y)
//     }
// }

impl<'a> ByteLaminationHelper<'a, Cow<'a, str>> for &'a str {
    fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]> {
        Cow::Borrowed(self.as_bytes())
    }
}

impl<'a> ByteLaminationHelper<'a, String> for &'a str {
    fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]> {
        Cow::Borrowed(self.as_bytes())
    }
}

impl<'a> ByteLaminationHelper<'a, Cow<'a, [u8]>> for &'a [u8] {
    fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]> {
        Cow::Borrowed(&self)
    }
}

impl<'a> ByteLaminationHelper<'a, Vec<u8>> for &'a [u8] {
    fn as_cow_bytes_helper(&self) -> Cow<'_, [u8]> {
        Cow::Borrowed(&self)
    }
}
